#!/usr/bin/env python

import sys
import socket
import utility
from optparse import OptionParser
from redux.pipelines import StandardPipeline

class Client(object):
	def process_request(self, request):
		raise NotImplementedError()
	def process_kwargs(self, **kwargs):
		raise NotImplementedError()

class NetworkClient(Client):
	def __init__(self, host, port=utility.REDUX_PORT):
		self.host = host
		self.port = port

	def process_request(self, request):
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		sock.connect((self.host, self.port))
		wfile = sock.makefile('wb')
		rfile = sock.makefile('rb')
		wfile.write(request)
		wfile.write('\n')
		wfile.flush()
		response = rfile.read()
		sock.close()
		return response

	def process_kwargs(self, **kwargs):
		request = utility.kwargs_to_request(**kwargs)
		return self.process_request(request)

class SimpleClient(Client):
	def process_request(self, request):
		kwargs = utility.request_to_kwargs(request)
		return self.process_kwargs(**kwargs)

	def process_kwargs(self, **kwargs):
		return StandardPipeline().process(**kwargs)

def add_option(parser, optname, help):
			dest = optname
			arg = '--%s' % (dest,)
			parser.add_option(arg, action='store', type='string', dest=dest, help=help)

def parse_argv():
	parser = OptionParser()
	for pipe in StandardPipeline.pipeorder:
		pipe_name = pipe.__name__
		if pipe.switch_arg:
			help = '%s: boolean switch to turn it on' % (pipe_name,)
			add_option(parser, pipe.switch_arg, help)
		if pipe.required_args:
			for arg in pipe.required_args:
				help = '%s: required' % (pipe_name,)
				add_option(parser, arg, help)
		if pipe.optional_args:
			for arg in pipe.optional_args:
				help = '%s: optional' % (pipe_name,)
				add_option(parser, arg, help)

	add_option(parser, 'request', 'full request as a single URL option string')
	add_option(parser, 'server_host', 'redux server host name (if not given, will start built-in redux processor)')
	add_option(parser, 'server_port', 'redux server port (if not given, will use default port)')

	(options, args) = parser.parse_args()


	kwargs = {}
	for key,value in options.__dict__.items():
		if value is not None:
			kwargs[key] = value
	if not kwargs:
		parser.print_help(sys.stderr)
		sys.exit(0)
	return kwargs

def run():
	kwargs = parse_argv()
	if 'server_host' in kwargs and kwargs['server_host']:
		if 'server_port' in kwargs and kwargs['server_port']:
			port = int(kwargs['server_port'])
		else:
			port = utility.REDUX_PORT
		client = NetworkClient(kwargs['server_host'], port)
	else:
		client = SimpleClient()

	if 'request' in kwargs and kwargs['request']:
		result = client.process_request(kwargs['request'])
	else:
		result = client.process_kwargs(**kwargs)
		
	sys.stdout.write(result)

if __name__ == '__main__':
	run()
